import gc
import os
import secrets
import time
from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from nebu import Message, Processor, processor
from nebu.config import GlobalConfig as NebuGlobalConfig
from nebu.containers.models import V1EnvVar
from nebu.errors import RetriableError
from nebu.processors.models import (
    V1Scale,
    V1ScaleDown,
    V1ScaleUp,
    V1ScaleZero,
)
from pydantic import BaseModel

from orign import V1TrainingStatus, find_latest_checkpoint

BASE_MODEL_ID = "unsloth/Qwen2.5-VL-32B-Instruct"


class TrainingRequest(BaseModel):
    adapter: str
    dataset: str
    model: str = BASE_MODEL_ID
    max_length: int = 32_768
    epochs: int = 1
    batch_size: int = 2
    gradient_accumulation_steps: int = 4
    learning_rate: float = 2e-4
    weight_decay: float = 0.01
    warmup_steps: int = 5
    logging_steps: int = 1
    save_steps: int = 5
    lora_alpha: int = 128
    lora_rank: int = 64
    lora_dropout: float = 0
    optimizer: str = "adamw_8bit"
    owner: Optional[str] = None
    labels: Optional[Dict[str, str]] = None


class TrainingResponse(BaseModel):
    loss: float
    train_steps_per_second: float
    train_samples_per_second: float
    train_runtime: float
    adapter: str
    adapter_uri: str


scale = V1Scale(
    up=V1ScaleUp(above_pressure=10, duration="5m"),
    down=V1ScaleDown(below_pressure=2, duration="10m"),
    zero=V1ScaleZero(duration="10m"),
)

# pip install torch torchvision torchaudio --index-url https://download.pytorch.org/whl/cu126
# pip uninstall -y xformers
# pip install -U xformers --index-url https://download.pytorch.org/whl/cu126
setup_script = """
apt update
apt install -y git
pip install torchvision torchaudio --index-url https://download.pytorch.org/whl/cu128
pip install trl peft transformers bitsandbytes sentencepiece accelerate orign
pip install -e git+https://github.com/pbarker/unsloth-zoo.git#egg=unsloth_zoo
pip install unsloth
pip install -e git+https://github.com/pbarker/unsloth-zoo.git#egg=unsloth_zoo
pip install huggingface_hub[hf_xet]
"""
# pip install -e git+https://github.com/pbarker/unsloth.git#egg=unsloth


ADAPTER_DIR = "/nebu/cache/adapters"
os.makedirs(ADAPTER_DIR, exist_ok=True)


def init():
    import gc

    from unsloth import FastVisionModel  # type: ignore # isort: skip
    import torch  # type: ignore  # type: ignore
    from nebu import Cache  # type: ignore
    from peft import LoraConfig, PeftModel  # type: ignore
    from unsloth_zoo.peft_utils import get_peft_regex  # type: ignore

    from orign import V1Adapter

    if "state" in globals():  # <-- already loaded by an earlier worker
        print("state already loaded by an earlier worker")
        return

    gc.collect()
    torch.cuda.empty_cache()  # os.environ.setdefault("MAX_PIXELS", "100352")

    @dataclass
    class TrainingState:
        base_model: FastVisionModel
        model_processor: Any
        base_model_id: str
        adapters: List[V1Adapter]
        cache: Cache

    print("Loading base model and tokenizer...")
    base_model, model_processor = FastVisionModel.from_pretrained(
        BASE_MODEL_ID,
        dtype=torch.bfloat16,
        load_in_4bit=False,
        max_seq_length=32_768,
        use_gradient_checkpointing="unsloth",
    )
    print("Base model and tokenizer loaded.")

    print("\nApplying initial PEFT setup with FastVisionModel.get_peft_model...")
    plumbed_model: PeftModel = FastVisionModel.get_peft_model(
        base_model,
        r=64,
        lora_alpha=128,
        lora_dropout=0.0,
        bias="none",
        finetune_vision_layers=True,
        finetune_language_layers=True,
        # target_modules is determined internally by Unsloth based on above flags
    )
    print(f"Type of model after get_peft_model: {type(plumbed_model)}")

    global G_INITIAL_TARGET_MODULES_PATTERN

    # --- Capture the target_modules from the "default" adapter ---
    if "default" in plumbed_model.peft_config:
        G_INITIAL_TARGET_MODULES_PATTERN = plumbed_model.peft_config[
            "default"
        ].target_modules
        print(
            "Captured initial target_modules pattern from 'default' adapter's config."
        )

        print("Deleting 'default' adapter created by get_peft_model.")
        plumbed_model.delete_adapter("default")
    else:
        print(
            "Warning: 'default' adapter not found. Attempting to generate target_modules pattern manually."
        )
        G_INITIAL_TARGET_MODULES_PATTERN = get_peft_regex(
            base_model,
            finetune_vision_layers=True,
            finetune_language_layers=True,
            finetune_attention_modules=True,
            finetune_mlp_modules=True,
        )
        print("Generated initial target_modules pattern (fallback).")

    if G_INITIAL_TARGET_MODULES_PATTERN is None:
        raise RuntimeError(
            "Could not determine initial target_modules pattern. Aborting."
        )

    plumbed_model.active_adapter = None
    print(
        f"Initial target_modules pattern to be reused: '{str(G_INITIAL_TARGET_MODULES_PATTERN)[:200]}...'"
    )

    global state
    state = TrainingState(
        base_model=plumbed_model,
        model_processor=model_processor,
        base_model_id=BASE_MODEL_ID,
        adapters=[],
        cache=Cache(),
    )


def add_or_load_adapter_for_model(
    model: "PeftModel",  # type: ignore # noqa: F821
    adapter_name: str,
    resume_training: bool,
):
    from peft import LoraConfig, PeftModel  # type: ignore

    global G_INITIAL_TARGET_MODULES_PATTERN
    print(
        f"\n[Adapter Management] Called for adapter: '{adapter_name}', resume: {resume_training}"
    )

    # This is the base folder for the adapter (e.g., ./adapters/adapter_A)
    adapter_base_folder = os.path.join(ADAPTER_DIR, adapter_name)
    os.makedirs(adapter_base_folder, exist_ok=True)  # Ensure base folder exists

    # Based on logs (ls adapters/adapter_A/adapter_A/), PEFT saves into a nested subdirectory
    # if save_pretrained is called with a path like "./adapters/adapter_A" and adapter name "adapter_A".
    # So, the actual files (adapter_config.json, etc.) are expected in this nested path.
    path_containing_adapter_files = os.path.join(adapter_base_folder, adapter_name)

    if adapter_name not in model.peft_config:
        print(
            f"[Adapter Management] Adapter '{adapter_name}' not found in peft_config. Adding new adapter configuration."
        )

        new_lora_config = LoraConfig(
            r=64,
            lora_alpha=128,
            lora_dropout=1e-3,
            bias="none",
            target_modules=G_INITIAL_TARGET_MODULES_PATTERN,
        )
        try:
            model.add_adapter(adapter_name=adapter_name, peft_config=new_lora_config)
            print(
                f"[Adapter Management] Adapter '{adapter_name}' added with new config."
            )
        except Exception as e:
            print(
                f"[Adapter Management] Error during model.add_adapter for '{adapter_name}': {e}"
            )
            raise
    else:
        print(
            f"[Adapter Management] Adapter '{adapter_name}' already exists in peft_config."
        )

    if resume_training:
        print(
            f"[Adapter Management] Attempting to load weights for adapter '{adapter_name}'."
        )
        print(
            f"[Adapter Management] Expected adapter files location: {path_containing_adapter_files}"
        )
        if os.path.isdir(path_containing_adapter_files):
            print(
                f"[Adapter Management] Contents of '{path_containing_adapter_files}': {os.listdir(path_containing_adapter_files)}"
            )
            try:
                # Load from the path where adapter_config.json and weights are actually located
                model.load_adapter(
                    path_containing_adapter_files, adapter_name, is_trainable=True
                )
                print(
                    f"[Adapter Management] Successfully loaded weights for adapter '{adapter_name}' from '{path_containing_adapter_files}'."
                )
            except Exception as e:
                print(
                    f"[Adapter Management] Error loading adapter weights for '{adapter_name}' from '{path_containing_adapter_files}': {e}"
                )
        else:
            print(
                f"[Adapter Management] Path '{path_containing_adapter_files}' does not exist or is not a directory. Cannot load weights."
            )

    model.set_adapter(adapter_name)
    print(f"[Adapter Management] Active adapter(s) set to: '{model.active_adapters}'")
    # The folder returned is the base folder where the nested adapter specific folder is.
    # For saving, we'll save to adapter_base_folder, and PEFT will create the nested structure if it does so.
    return adapter_base_folder


def drop_adapter_from_model(model: "PeftModel", adapter_name_to_drop: str):  # type: ignore # noqa: F821
    import torch  # type: ignore

    print(
        f"\n[Adapter Management] Attempting to drop adapter: '{adapter_name_to_drop}'"
    )
    print(
        f"[Adapter Management] Current peft_config keys before deleting: {list(model.peft_config.keys())}"
    )
    print(
        f"[Adapter Management] Currently active adapters before deleting: {model.active_adapters}"
    )

    if adapter_name_to_drop in model.peft_config:
        try:
            print(
                f"[Adapter Management] Calling model.delete_adapter('{adapter_name_to_drop}')..."
            )
            model.delete_adapter(adapter_name_to_drop)
            print(
                f"[Adapter Management] Adapter '{adapter_name_to_drop}' deleted successfully from peft_config."
            )
        except Exception as e:
            print(
                f"[Adapter Management] Error during model.delete_adapter('{adapter_name_to_drop}'): {e}"
            )
    else:
        print(
            f"[Adapter Management] Adapter '{adapter_name_to_drop}' not found in peft_config. Cannot delete."
        )

    if not model.active_adapters:
        model.active_adapter = None
        print(
            "[Adapter Management] No active adapters remain. Set model.active_adapter to None."
        )
    else:
        print(
            f"[Adapter Management] Active adapters after drop attempt: {model.active_adapters}"
        )

    torch.cuda.empty_cache()
    gc.collect()
    print(f"[Adapter Management] Finished dropping adapter '{adapter_name_to_drop}'.\n")


def train_lora_adapter(
    adapter_name_to_train: str,
    training_dataset: Any,
    num_epochs: int = 1,
    resume_from_saved_state: bool = False,
    checkpoint_path: Optional[str] = None,
):
    import json  # Added for reading trainer_state.json
    import os  # ensure os is imported locally if not already

    import torch  # type: ignore
    from trl import SFTConfig, SFTTrainer  # type: ignore
    from unsloth import FastVisionModel, is_bf16_supported  # type: ignore
    from unsloth.trainer import UnslothVisionDataCollator  # type: ignore

    global state

    print(
        f"\n--- Starting train_lora_adapter for adapter: '{adapter_name_to_train}' (Epochs: {num_epochs}, Resume: {resume_from_saved_state}) ---"
    )

    # add_or_load_adapter_for_model now returns the base save folder (e.g., ./adapters/adapter_A)
    adapter_base_save_folder = add_or_load_adapter_for_model(
        state.base_model,
        adapter_name_to_train,
        resume_from_saved_state or bool(checkpoint_path),
    )
    print(
        f"Adapter base save folder for '{adapter_name_to_train}': {adapter_base_save_folder}"
    )

    print("\nPreparing model for training with FastVisionModel.for_training...")
    model_ready_for_training = FastVisionModel.for_training(state.base_model)
    print("Model prepared for training.")

    print("\nModel's trainable parameters (on instance passed to SFTTrainer):")
    try:
        model_ready_for_training.print_trainable_parameters()
    except AttributeError:
        total_params = sum(p.numel() for p in model_ready_for_training.parameters())
        trainable_params = sum(
            p.numel() for p in model_ready_for_training.parameters() if p.requires_grad
        )
        print(
            f"trainable params: {trainable_params:,} || all params: {total_params:,} || trainable%: {100 * trainable_params / total_params:.4f}"
        )

    initial_learning_rate = 4e-4  # 5e-5
    print(f"Using initial learning_rate for SFTTrainer: {initial_learning_rate}")

    sft_config_args = SFTConfig(
        per_device_train_batch_size=1,
        gradient_accumulation_steps=1,
        num_train_epochs=num_epochs,
        learning_rate=initial_learning_rate,
        optim="adamw_8bit",
        fp16=not is_bf16_supported(),
        bf16=is_bf16_supported(),
        save_strategy="epoch",
        save_total_limit=1,
        output_dir=f"./runs/{adapter_name_to_train}",
        logging_steps=1,
        report_to="none",
    )

    trainer = SFTTrainer(
        model=model_ready_for_training,
        tokenizer=state.model_processor,
        data_collator=UnslothVisionDataCollator(
            model_ready_for_training, state.model_processor, resize="max"
        ),
        train_dataset=training_dataset,
        args=sft_config_args,
    )

    # For SFTTrainer's resume_from_checkpoint:
    # If checkpoint_path is explicitly given, use it.
    # Else, if resume_from_saved_state is True, pass True to trainer.train() to load latest from output_dir.
    # Otherwise, no resume.
    sft_trainer_resume_arg = None
    if checkpoint_path:
        sft_trainer_resume_arg = checkpoint_path
        print(
            f"SFTTrainer will attempt to resume from EXPLICIT checkpoint path: '{sft_trainer_resume_arg}'"
        )
    elif resume_from_saved_state:
        sft_trainer_resume_arg = (
            True  # Let Trainer find the latest checkpoint in output_dir
        )
        print(
            f"SFTTrainer will attempt to resume from the latest checkpoint in its output_dir: {sft_config_args.output_dir}"
        )
    else:
        print(
            "SFTTrainer training from scratch (no SFTTrainer checkpoint specified or found for resume)."
        )

    # Check if the directory for SFTTrainer resume exists if a path was constructed (not True)
    if isinstance(sft_trainer_resume_arg, str) and not os.path.isdir(
        sft_trainer_resume_arg
    ):
        print(
            f"Warning: SFTTrainer resume path '{sft_trainer_resume_arg}' not found. Training from scratch."
        )
        sft_trainer_resume_arg = None  # Fallback to no resume

    print("\nInspecting SFT checkpoint for prior epoch count before training starts...")
    sft_checkpoint_dir_to_inspect = None
    if isinstance(sft_trainer_resume_arg, str) and os.path.isdir(
        sft_trainer_resume_arg
    ):
        sft_checkpoint_dir_to_inspect = sft_trainer_resume_arg
        print(
            f"SFTTrainer is configured to resume from explicit path: {sft_checkpoint_dir_to_inspect}"
        )
    elif sft_trainer_resume_arg is True:
        # SFTTrainer will look in sft_config_args.output_dir
        print(
            f"SFTTrainer is configured to resume from latest in output_dir: {sft_config_args.output_dir}"
        )
        # We need to find what SFTTrainer would find.
        # find_latest_checkpoint is imported from orign, which should be available.
        latest_sft_checkpoint_in_output_dir = find_latest_checkpoint(
            sft_config_args.output_dir
        )
        if latest_sft_checkpoint_in_output_dir:
            sft_checkpoint_dir_to_inspect = latest_sft_checkpoint_in_output_dir
            print(
                f"Found latest SFT checkpoint for inspection: {sft_checkpoint_dir_to_inspect}"
            )
        else:
            print(
                f"No SFT checkpoint found in {sft_config_args.output_dir} to inspect for prior epochs."
            )
    else:
        print(
            "SFTTrainer is not configured to resume from a checkpoint. No prior SFT epochs to report from trainer_state.json."
        )

    if sft_checkpoint_dir_to_inspect:
        trainer_state_path_to_inspect = os.path.join(
            sft_checkpoint_dir_to_inspect, "trainer_state.json"
        )
        print(
            f"Attempting to read SFT trainer state from: {trainer_state_path_to_inspect}"
        )
        if os.path.exists(trainer_state_path_to_inspect):
            try:
                with open(trainer_state_path_to_inspect, "r") as f:
                    sft_state_data = json.load(f)
                sft_epochs_completed = sft_state_data.get("epoch", 0.0)
                sft_global_step = sft_state_data.get("global_step", 0)
                print(
                    f"  >> SFT Checkpoint State: Epochs completed = {sft_epochs_completed}, Global steps = {sft_global_step}"
                )
            except Exception as e:
                print(
                    f"  >> Warning: Failed to read/parse SFT {trainer_state_path_to_inspect}: {e}"
                )
        else:
            print(f"  >> Warning: SFT {trainer_state_path_to_inspect} not found.")
    print("--- End of SFT Checkpoint Inspection ---")

    print("\nStarting SFTTrainer training...")
    trainer.train(resume_from_checkpoint=sft_trainer_resume_arg)
    print("SFTTrainer training finished.")

    # Save adapter weights to the base folder. PEFT might create a nested folder.
    # Example: if adapter_base_save_folder is "./adapters/adapter_A", PEFT saves to "./adapters/adapter_A/adapter_A/"
    # This matches the loading logic.
    print(
        f"\nSaving adapter weights for '{adapter_name_to_train}' to base folder: {adapter_base_save_folder}"
    )
    model_ready_for_training.save_pretrained(adapter_base_save_folder)
    print("Adapter weights saved.")

    drop_adapter_from_model(state.base_model, adapter_name_to_train)

    del trainer, model_ready_for_training
    torch.cuda.empty_cache()
    gc.collect()
    print(
        f"--- train_lora_adapter for adapter: '{adapter_name_to_train}' completed ---\n"
    )


def train_unsloth_sft(message: Message[TrainingRequest]) -> TrainingResponse:
    import gc
    import json
    import shutil

    import requests
    import torch  # type: ignore
    from chatmux import oai_to_unsloth
    from nebu import (
        Bucket,
        ContainerConfig,
        V1ResourceReference,
        is_allowed,
    )

    from orign import Adapter, Training, V1LoraParams

    global state
    if not hasattr(state, "base_model") or not hasattr(state, "model_processor"):
        raise RuntimeError(
            "Base model and processor not initialized. Ensure init() has run."
        )

    # First ensure CUDA cache is cleared
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
        torch.cuda.ipc_collect()

    # Force garbage collection multiple times to ensure all tensors are released
    gc.collect()

    print("message", message)
    if not message.content:
        raise ValueError("No training request provided")
    training_request: TrainingRequest = message.content

    container_config = ContainerConfig.from_env()
    print("container_config", container_config)

    bucket = Bucket()

    print("determining adapter namespace and name...")
    adapter_parts = training_request.adapter.split("/")
    if len(adapter_parts) == 2:
        print("adapter_parts", adapter_parts)
        adapter_namespace = adapter_parts[0]
        adapter_name = adapter_parts[1]
    else:
        adapter_name = training_request.adapter
        if training_request.owner:
            print("owner", training_request.owner)
            adapter_namespace = training_request.owner
        else:
            print("no owner, using message.handle", message.handle)
            adapter_namespace = message.handle

    print("adapter_namespace", adapter_namespace)
    if not adapter_namespace:
        raise ValueError("Could not determine adapter namespace")

    # Define local and bucket paths
    # ADAPTER_DIR is global: /nebu/cache/adapters
    local_adapter_weights_dir_for_current_adapter = os.path.join(
        ADAPTER_DIR, adapter_name
    )
    # train_lora_adapter saves adapter weights into a subfolder named adapter_name within this
    actual_local_adapter_files_path = os.path.join(
        local_adapter_weights_dir_for_current_adapter, adapter_name
    )

    local_sft_runs_dir = f"./runs/{adapter_name}"

    # Bucket URIs
    adapter_weights_bucket_uri = f"{container_config.namespace_volume_uri}/adapters/{adapter_namespace}/{adapter_name}"
    sft_checkpoints_bucket_uri = f"{container_config.namespace_volume_uri}/sft_runs/{adapter_namespace}/{adapter_name}"

    if training_request.labels:
        training_labels = training_request.labels.copy()
    else:
        training_labels = {}
    training_labels["message_id"] = message.id
    training_labels["container_id"] = os.getenv("NEBU_CONTAINER_ID", "unknown")
    random_chars = secrets.token_hex(3)

    adapter_ref = V1ResourceReference(
        name=adapter_name,
        namespace=adapter_namespace,
        kind="Adapter",
    )
    print("adapter_ref: ", adapter_ref)

    training = None
    try:
        print("creating training with api_key", message.api_key)
        training = Training(
            name=adapter_name + "-" + random_chars,
            namespace=adapter_namespace,
            config_data=message.model_dump(),
            adapter=adapter_ref,
            labels=training_labels,
            unique_adapter_active=True,
            api_key=message.api_key,
        )
        print("\n >> marking initial training as running")
        training.update(status=V1TrainingStatus.RUNNING)
    except Exception as e:
        print(
            f"FATAL: Failed to create or update Training resource for {adapter_ref}: {e}  --- retrying..."
        )
        raise RetriableError(
            f"Failed to set up Training resource: {e}  --- retrying..."
        ) from e

    failure = False
    try:
        adapter_resource = None
        try:
            adapters_found = Adapter.get(
                adapter_namespace, adapter_name, api_key=message.api_key
            )
            if adapters_found:
                adapter_resource = adapters_found[0]
        except Exception:
            adapters_found = []  # noqa
        print("found adapter resource", adapter_resource)

        is_continue = False
        epochs_trained_so_far = 0
        sft_checkpoint_to_resume_from = (
            None  # This will be local_sft_runs_dir if resuming
        )

        if adapter_resource:
            print("Found existing adapter resource: ", adapter_resource)
            epochs_trained_so_far = adapter_resource.epochs_trained

            if not is_allowed(
                adapter_resource.metadata.owner, message.user_id, message.orgs
            ):
                raise ValueError("You are not allowed to train this existing adapter")

            # Sync adapter weights from bucket to local ADAPTER_DIR/adapter_name
            # train_lora_adapter expects them in ADAPTER_DIR/adapter_name/adapter_name
            print(
                f"Attempting to sync adapter weights from {adapter_weights_bucket_uri} to {actual_local_adapter_files_path}"
            )
            os.makedirs(
                actual_local_adapter_files_path, exist_ok=True
            )  # Ensure target dir exists for sync
            try:
                bucket.sync(adapter_weights_bucket_uri, actual_local_adapter_files_path)
                print(f"Synced adapter weights to {actual_local_adapter_files_path}")
                is_continue = True
            except Exception as e:
                print(
                    f"Warning: Failed to sync adapter weights from {adapter_weights_bucket_uri}: {e}. May proceed without them if adapter is being created fresh by train_lora_adapter."
                )

            # Check if we have a specific checkpoint URI instead of a general SFT runs directory
            checkpoint_uri = adapter_resource.checkpoint_uri
            # If checkpoint URI points to a specific checkpoint (contains "/checkpoint-")
            if checkpoint_uri and "/checkpoint-" in checkpoint_uri:
                checkpoint_name = os.path.basename(checkpoint_uri)
                print(
                    f"Found specific checkpoint reference: {checkpoint_name} in {checkpoint_uri}"
                )

                # Create local directory for this specific checkpoint
                local_checkpoint_dir = os.path.join(local_sft_runs_dir, checkpoint_name)
                os.makedirs(local_checkpoint_dir, exist_ok=True)

                # Sync that specific checkpoint from bucket
                try:
                    print(
                        f"Syncing specific checkpoint from {checkpoint_uri} to {local_checkpoint_dir}"
                    )
                    bucket.sync(checkpoint_uri, local_checkpoint_dir)
                    print(f"Successfully synced checkpoint {checkpoint_name}")
                    sft_checkpoint_to_resume_from = local_checkpoint_dir
                    is_continue = True
                except Exception as e:
                    print(f"Failed to sync specific checkpoint: {e}")
                    is_continue = False
            else:
                # Fallback to old behavior in case checkpoint_uri is not a specific checkpoint
                print(
                    f"No specific checkpoint reference found in adapter. Will sync from general checkpoint directory: {checkpoint_uri}"
                )
                try:
                    bucket.sync(checkpoint_uri, local_sft_runs_dir)
                    # Look for latest checkpoint in the synced directory
                    latest_checkpoint = find_latest_checkpoint(local_sft_runs_dir)
                    if latest_checkpoint:
                        print(
                            f"Found latest checkpoint in synced directory: {latest_checkpoint}"
                        )
                        sft_checkpoint_to_resume_from = latest_checkpoint
                        is_continue = True
                    else:
                        print("No valid checkpoint found in synced directory")
                        is_continue = False
                except Exception as e:
                    print(f"Failed to sync from general checkpoint directory: {e}")
                    is_continue = False
        else:
            print(
                f"No existing Adapter resource found for {adapter_namespace}/{adapter_name}. This will be a new training."
            )
            # Clean up local directories for a fresh start if they exist from a failed previous run
            if os.path.exists(local_adapter_weights_dir_for_current_adapter):
                shutil.rmtree(local_adapter_weights_dir_for_current_adapter)
                print(
                    f"Cleaned up existing local adapter dir: {local_adapter_weights_dir_for_current_adapter}"
                )
            if os.path.exists(local_sft_runs_dir):
                shutil.rmtree(local_sft_runs_dir)
                print(f"Cleaned up existing local SFT runs dir: {local_sft_runs_dir}")
            os.makedirs(local_adapter_weights_dir_for_current_adapter, exist_ok=True)
            os.makedirs(local_sft_runs_dir, exist_ok=True)

        print("Downloading dataset")
        time_start_download = time.time()
        response = requests.get(training_request.dataset)
        response.raise_for_status()
        print(f"Downloaded dataset in {time.time() - time_start_download} seconds")

        lines = response.content.decode("utf-8").splitlines()
        time_start_convert = time.time()
        converted_dataset = [
            oai_to_unsloth(json.loads(line)) for line in lines if line.strip()
        ]
        print(f"Converted dataset in {time.time() - time_start_convert} seconds")
        print("dataset example", converted_dataset[:1])

        # Calculate the cumulative target number of epochs
        cumulative_target_epochs = epochs_trained_so_far + training_request.epochs
        print(
            f"Adapter '{adapter_name}': Has {epochs_trained_so_far} epochs trained. Requesting additional {training_request.epochs} epochs. Target cumulative epochs: {cumulative_target_epochs}."
        )
        print(
            f"Calling train_lora_adapter for '{adapter_name}'. Resume SFT: {is_continue}, SFT checkpoint path hint: {sft_checkpoint_to_resume_from}"
        )

        time_start_train = time.time()
        train_lora_adapter(
            adapter_name_to_train=adapter_name,
            training_dataset=converted_dataset,
            num_epochs=cumulative_target_epochs,  # Pass cumulative target epochs
            resume_from_saved_state=is_continue,  # For adapter weights loading
            checkpoint_path=sft_checkpoint_to_resume_from
            if is_continue
            else None,  # For SFTTrainer resume
        )
        print(
            f"train_lora_adapter completed in {time.time() - time_start_train} seconds"
        )

        # After training, sync artifacts to bucket
        # 1. Sync adapter weights
        # train_lora_adapter saves them to actual_local_adapter_files_path
        if os.path.exists(actual_local_adapter_files_path) and os.listdir(
            actual_local_adapter_files_path
        ):
            print(
                f"Syncing adapter weights from {actual_local_adapter_files_path} to {adapter_weights_bucket_uri}"
            )
            bucket.copy(actual_local_adapter_files_path, adapter_weights_bucket_uri)
            print("Synced adapter weights to bucket.")
        else:
            print(
                f"Warning: Local adapter files path {actual_local_adapter_files_path} is empty or does not exist after training. Cannot sync to bucket."
            )

        # 2. Sync SFT checkpoints
        # train_lora_adapter's SFTTrainer saves to local_sft_runs_dir
        if os.path.exists(local_sft_runs_dir) and os.listdir(local_sft_runs_dir):
            # Find the latest checkpoint directory
            latest_checkpoint = find_latest_checkpoint(local_sft_runs_dir)
            if latest_checkpoint:
                # Extract the checkpoint name (e.g., "checkpoint-132")
                latest_checkpoint_name = os.path.basename(latest_checkpoint)
                # Create a specific path for just the latest checkpoint
                latest_checkpoint_bucket_uri = (
                    f"{sft_checkpoints_bucket_uri}/{latest_checkpoint_name}"
                )

                print(
                    f"Syncing latest SFT checkpoint from {latest_checkpoint} to {latest_checkpoint_bucket_uri}"
                )
                bucket.copy(latest_checkpoint, latest_checkpoint_bucket_uri)
                print(f"Synced latest checkpoint ({latest_checkpoint_name}) to bucket.")

                # We'll use this specific checkpoint URI instead of the parent directory
                checkpoint_uri_for_adapter = latest_checkpoint_bucket_uri
            else:
                print(
                    f"Warning: Could not find a checkpoint directory in {local_sft_runs_dir}"
                )
                checkpoint_uri_for_adapter = sft_checkpoints_bucket_uri
        else:
            print(
                f"Warning: Local SFT runs directory {local_sft_runs_dir} is empty or does not exist after training. Cannot sync to bucket."
            )
            checkpoint_uri_for_adapter = sft_checkpoints_bucket_uri

        # Collect metrics from trainer_state.json in the local SFT runs directory
        training_metrics = {}
        trainer_state_path = None  # Initialize to None

        latest_checkpoint_dir = find_latest_checkpoint(local_sft_runs_dir)
        if latest_checkpoint_dir:
            trainer_state_path = os.path.join(
                latest_checkpoint_dir, "trainer_state.json"
            )
            print(f"Found latest checkpoint directory: {latest_checkpoint_dir}")
            print(f"Attempting to load trainer_state.json from: {trainer_state_path}")
        else:
            print(
                f"Warning: No checkpoint directory found in {local_sft_runs_dir}. Looking for trainer_state.json in the root of SFT runs dir as a fallback."
            )
            # Fallback to old behavior if no checkpoint dir is found, though less likely to succeed
            trainer_state_path = os.path.join(local_sft_runs_dir, "trainer_state.json")

        if trainer_state_path and os.path.exists(trainer_state_path):
            try:
                with open(trainer_state_path, "r") as f:
                    state_data = json.load(f)
                training_metrics = state_data
                print(f"Loaded training metrics from {trainer_state_path}")
                log_history = state_data.get("log_history", [])
                if log_history:
                    print(f"  Final log_history entry: {log_history[-1]}")
            except Exception as e:
                print(f"Warning: Failed to read/parse {trainer_state_path}: {e}.")
        else:
            print(f"Warning: {trainer_state_path} not found. Metrics will be empty.")

        # Update Adapter resource with the specific checkpoint URI
        final_epochs_trained = epochs_trained_so_far + training_request.epochs

        Adapter(
            name=adapter_name,
            namespace=adapter_namespace,
            model_uri=adapter_weights_bucket_uri,  # URI now points to the adapter weights in the bucket
            checkpoint_uri=checkpoint_uri_for_adapter,  # URI now points to the specific latest checkpoint
            owner=training_request.owner or message.user_id,  # type: ignore
            base_model=training_request.model,  # This is the original base model ID like "unsloth/Qwen2.5-VL-32B-Instruct"
            epochs_trained=final_epochs_trained,
            examples_trained=(
                adapter_resource.examples_trained if adapter_resource else 0
            )
            + len(converted_dataset),
            last_trained=int(time.time()),
            lora=V1LoraParams(
                r=training_request.lora_rank,
                alpha=training_request.lora_alpha,
                dropout=training_request.lora_dropout,
            ),
            labels=training_request.labels,
            api_key=message.api_key,
        )

        training.log(data=training_metrics)

        gc.collect()
        if torch.cuda.is_available():
            torch.cuda.empty_cache()
            torch.cuda.ipc_collect()

        # Extract metrics for TrainingResponse
        # Use .get() with defaults to avoid KeyError if metrics are missing
        final_loss = 0.0
        train_steps_per_second = 0.0
        train_samples_per_second = 0.0
        train_runtime = 0.0

        if training_metrics:
            log_history = training_metrics.get("log_history", [])
            if log_history:  # Get loss from the last step
                final_loss = log_history[-1].get(
                    "loss", log_history[-1].get("train_loss", 0.0)
                )

            # These specific keys might not be in log_history but directly in trainer_stats from SFTTrainer
            # train_lora_adapter doesn't directly return trainer_stats, so we rely on trainer_state.json
            # which might have slightly different structure for aggregated stats.
            # For now, let's use what's typically available or default to 0.
            train_steps_per_second = training_metrics.get(
                "train_steps_per_second", 0.0
            )  # This key might not exist directly
            train_samples_per_second = training_metrics.get(
                "train_samples_per_second", 0.0
            )  # This key might not exist
            train_runtime = training_metrics.get(
                "train_runtime", time.time() - time_start_train
            )  # Fallback to measured time

        return TrainingResponse(
            loss=final_loss,
            train_steps_per_second=train_steps_per_second,
            train_samples_per_second=train_samples_per_second,
            train_runtime=train_runtime,
            adapter=training_request.adapter,
            adapter_uri=adapter_weights_bucket_uri,  # Return the bucket URI for the adapter weights
        )
    except Exception as e:
        print(f"Error training unsloth: {e}")
        failure = True
        if training:
            print("\n >> marking training as failed due to exception")
            training.update(status=V1TrainingStatus.FAILED)
        # If an error occurs, we must still return a TrainingResponse or raise.
        # Raising the original error is often better for debugging.
        raise
    finally:
        print(
            f"finally block: training resource exists: {bool(training)}, failure: {failure}"
        )
        if training:
            if failure:
                if (
                    training.training.status != V1TrainingStatus.FAILED
                ):  # Avoid double update if already set
                    print("\n >> ensuring training is marked as FAILED in finally")
                    training.update(status=V1TrainingStatus.FAILED)
            else:
                if training.training.status != V1TrainingStatus.COMPLETED:
                    print("\n >> marking training as COMPLETED in finally")
                    training.update(status=V1TrainingStatus.COMPLETED)


def UnslothSFT(
    platform: str = "runpod",
    accelerators: List[str] = ["1:H100_SXM"],
    image: str = "us-docker.pkg.dev/agentsea-dev/orign/unsloth-train:latest",  # "runpod/pytorch:2.8.0-py3.11-cuda12.8.1-cudnn-devel-ubuntu22.04",  # "pytorch/pytorch:2.7.0-cuda12.8-cudnn9-devel" "pytorch/pytorch:2.6.0-cuda12.6-cudnn9-devel",
    scale: V1Scale = scale,
    namespace: Optional[str] = None,
    env: Optional[List[V1EnvVar]] = None,
    config: Optional[NebuGlobalConfig] = None,
    hot_reload: bool = True,
    debug: bool = False,
    min_replicas: int = 1,
    max_replicas: int = 4,
) -> Processor[TrainingRequest, TrainingResponse]:
    decorate = processor(
        image=image,
        # setup_script=setup_script,
        accelerators=accelerators,
        platform=platform,
        scale=scale,
        namespace=namespace,
        env=env,
        init_func=init,
        # execution_mode="subprocess",
        config=config,
        hot_reload=hot_reload,
        debug=debug,
        min_replicas=min_replicas,
        max_replicas=max_replicas,
    )
    return decorate(train_unsloth_sft)



from typing import List, Optional, Tuple

from PIL import Image
from pydantic import BaseModel

from .img import convert_images


class V1EnvState(BaseModel):
    """The state of the environment"""

    images: Optional[List[str]] = None
    coordinates: Optional[Tuple[int, int]] = None
    video: Optional[str] = None
    text: Optional[str] = None
    timestamp: Optional[float] = None


class EnvState:
    """The state of the environment"""

    def __init__(
        self,
        images: Optional[List[str | Image.Image | None]] = None,
        coordinates: Optional[Tuple[int, int]] = None,
        video: Optional[str] = None,
        text: Optional[str] = None,
        timestamp: Optional[float] = None,
    ) -> None:
        self.images = convert_images(images) if images else None  # type: ignore
        self.coordinates = coordinates
        self.video = video
        self.text = text
        self.timestamp = timestamp

    def to_v1(self) -> V1EnvState:
        return V1EnvState(
            images=self.images,
            coordinates=self.coordinates,
            video=self.video,
            text=self.text,
            timestamp=self.timestamp,
        )

    @classmethod
    def from_v1(cls, data: V1EnvState) -> "EnvState":
        out = cls.__new__(cls)
        out.images = convert_images(data.images) if data.images else None
        out.coordinates = data.coordinates
        out.video = data.video
        out.text = data.text
        out.timestamp = data.timestamp

        return out



import os
from typing import Dict, List, Optional

from nebu.containers.models import (
    V1ContainerRequest,
    V1EnvVar,
    V1Meter,
    V1ResourceMetaRequest,
    V1VolumeDriver,
    V1VolumePath,
)


class TRLRequest(V1ContainerRequest):
    """
    A wrapper class for creating TRL training jobs.
    """

    def __init__(
        self,
        name: str,
        model: str,
        platform: str,
        bucket: str,
        accelerators: List[str],
        namespace: Optional[str] = None,
        train_type: str = "sft",
        num_train_epochs: int = 1,
        save_steps: int = 1,
        save_total_limit: int = 3,
        save_strategy: str = "steps",
        dtype: str = "bfloat16",
        per_device_train_batch_size: int = 1,
        per_device_eval_batch_size: int = 1,
        output_dir: str = "/output",
        use_peft: bool = True,
        price_per_training_second: Optional[float] = None,
        max_seq_length: int = 2048,
        timeout: str = "2h",
        bucket_base_key: str = "trl",
        labels: Optional[Dict[str, str]] = None,
    ):
        """
        Initialize a TRL container request for transformer reinforcement learning.
        Inherits from V1ContainerRequest.

        Args:
            name: Name for the training job
            model: HuggingFace model ID to use
            platform: The compute platform to use
            bucket: S3 bucket for model storage
            accelerators: List of accelerators to use (e.g., ["A100"])
            namespace: Kubernetes namespace
            train_type: Type of training (e.g., "sft", "rl", "dpo")
            num_train_epochs: Number of training epochs
            save_steps: How often to save the model
            save_total_limit: Maximum number of checkpoints to keep
            save_strategy: When to save checkpoints
            dtype: Data type for training
            per_device_train_batch_size: Training batch size per device
            per_device_eval_batch_size: Evaluation batch size per device
            output_dir: Directory to save model outputs
            use_peft: Whether to use Parameter-Efficient Fine-Tuning
            price_per_training_second: Optional cost per training second
            max_seq_length: Maximum sequence length for training
            timeout: Timeout duration for the job
            bucket_base_key: Base key for S3 bucket path
            labels: Optional labels to add to resources
        """
        train_command = f"""
DATASET_PATH="{output_dir}/dataset.json"
echo "Downloading dataset to $DATASET_PATH"
wget "$DATASET_URI" -O "$DATASET_PATH"
source activate trl && trl {train_type} \
    --model_name_or_path $MODEL \
    --dataset_name $DATASET_PATH \
    --dataset_train_split "train" \
    --dataset_test_split "test" \
    --output_dir {output_dir} \
    --torch_dtype {dtype} \
    --max_seq_length {max_seq_length} \
    --per_device_train_batch_size {per_device_train_batch_size} \
    --per_device_eval_batch_size {per_device_eval_batch_size} \
    --use_peft {use_peft} \
    --save_strategy {save_strategy} \
    --save_steps {save_steps} \
    --save_total_limit {save_total_limit} \
    --num_train_epochs {num_train_epochs}
LATEST="$(ls -1d {output_dir}/checkpoint-* | sort -V | tail -n 1)"
rclone sync "$LATEST" "s3://{bucket}/{bucket_base_key}/{namespace}/{name}/latest"
        """

        train_meters = None
        if price_per_training_second:
            train_meters = [
                V1Meter(
                    cost=price_per_training_second,
                    unit="second",
                    metric="runtime",
                    currency="USD",
                )
            ]

        train_queue_name = f"trl-{name}"

        train_env = [
            V1EnvVar(key="MODEL", value=model),
        ]

        if os.getenv("HF_TOKEN"):
            train_env.append(V1EnvVar(key="HF_TOKEN", value=os.getenv("HF_TOKEN")))

        train_volumes = [
            V1VolumePath(
                source=f"{output_dir}",
                dest=f"s3://{bucket}/{bucket_base_key}/{namespace}/{name}/jobs/$NEBU_CONTAINER_ID",
                driver=V1VolumeDriver.RCLONE_SYNC,
                continuous=True,
            )
        ]

        # Initialize the parent V1ContainerRequest
        super().__init__(
            image="huggingface/trl-latest-gpu:latest",
            platform=platform,
            metadata=V1ResourceMetaRequest(
                name=name,
                namespace=namespace,
                labels=labels,
            ),
            command=train_command,
            accelerators=accelerators,
            env=train_env,
            volumes=train_volumes,
            meters=train_meters,
            restart="Never",
            queue=train_queue_name,
            timeout=timeout,
        )



import time
from typing import (
    Any,
    ClassVar,
    Dict,
    Generic,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    get_args,
)

import requests
from chatmux import ChatRequest
from nebu import V1ResourceMetaRequest, V1ResourceReference
from nebu.logging import logger
from pydantic import BaseModel, TypeAdapter

from orign.buffers.models import V1ReplayBufferData, V1Trainer
from orign.config import GlobalConfig
from orign.llms.models import (
    V1GenerateRequest,
    V1OnlineLLM,
    V1OnlineLLMRequest,
    V1OnlineLLMs,
    V1OnlineLLMStatus,
    V1UpdateOnlineLLMRequest,
)
from orign.trainings.models import V1TrainingStatus
from orign.trainings.training import Training

InputType = TypeVar("InputType", bound=BaseModel)
OutputType = TypeVar("OutputType", bound=BaseModel)
ExampleType = TypeVar("ExampleType", bound=BaseModel)


class OnlineLLM(Generic[InputType, OutputType, ExampleType]):
    """
    Online LLMs can both learn and act.
    """

    # Use ClassVar to indicate this is intended as a class-level attribute
    # The type hint still uses the TypeVar OutputType
    output_model: ClassVar[Type[OutputType]]  # type: ignore

    def __init_subclass__(cls, **kwargs):  # type: ignore
        """
        Automatically called when a class inherits from OnlineLLM.
        It inspects the base classes to find the parameterized
        OnlineLLM generic and extracts the OutputType.
        """
        # Always call super() for cooperative multiple inheritance
        super().__init_subclass__(**kwargs)
        logger.debug("Initializing subclass {}...", cls.__name__)

        # Get the original base classes tuple if it exists
        orig_bases = getattr(cls, "__orig_bases__", None)
        logger.debug("Original bases for {}: {}", cls.__name__, orig_bases)
        if not orig_bases:
            # This might happen if the class is not directly inheriting
            # a parameterized generic, or in edge cases.
            # We might only want to enforce this for direct subclasses.
            if OnlineLLM in cls.__bases__:
                logger.warning(
                    f"Subclass {cls.__name__} inherits from OnlineLLM but "
                    f"could not find original parameterized base."
                )
            return  # Cannot determine types

        # Find the specific OnlineLLM[...] base class among the original bases
        online_llm_base = None
        for base in orig_bases:
            origin = getattr(base, "__origin__", None)
            logger.debug("Checking base origin: {}", origin)
            if origin is OnlineLLM and hasattr(base, "__args__"):
                logger.debug("Found parameterized OnlineLLM base: {}", base)
                online_llm_base = base
                break

        if online_llm_base:
            args = get_args(online_llm_base)
            logger.debug("Type arguments for {}: {}", cls.__name__, args)
            # Expected order: InputType, OutputType, ExampleType
            if len(args) >= 2:
                output_type_arg = args[1]
                # Check if it's a concrete type (not a TypeVar)
                # and preferably a BaseModel (matching the bound)
                if isinstance(output_type_arg, type) and issubclass(
                    output_type_arg, BaseModel
                ):
                    logger.debug(
                        "Setting {}.output_model = {}", cls.__name__, output_type_arg
                    )
                    cls.output_model = output_type_arg  # type: ignore
                elif isinstance(output_type_arg, TypeVar):
                    logger.warning(
                        f"Output type for {cls.__name__} is still a TypeVar ({output_type_arg}). "
                        f"Subclass must provide a concrete type."
                    )
                else:
                    logger.warning(
                        f"Inferred output type {output_type_arg} for {cls.__name__} "
                        f"is not a concrete pydantic BaseModel."
                    )
                    # Decide if you want to assign anyway or raise an error
                    # cls.output_model = output_type_arg # Assigning might lead to runtime errors later
            else:
                logger.warning(
                    f"Parameterized base {online_llm_base} for {cls.__name__} "
                    f"does not have enough type arguments."
                )
        elif OnlineLLM in cls.__bases__:
            # Only raise error/warning if it's a *direct* subclass lacking parameterization
            raise TypeError(
                f"Class {cls.__name__} must inherit from a parameterized version of OnlineLLM, "
                f"e.g., OnlineLLM[MyInput, MyOutput, MyExample]"
            )

    def __init__(
        self,
        name: str,
        model: str,
        server: V1ResourceReference,
        trainer: V1Trainer,
        train_every: Optional[int] = None,
        sample_n: Optional[int] = None,
        sample_strategy: Optional[str] = None,
        num_epochs: Optional[int] = None,
        namespace: Optional[str] = None,
        labels: Optional[Dict[str, str]] = None,
        adapter: Optional[str] = None,
        config: Optional[GlobalConfig] = None,
        no_delete: bool = False,
        api_key: Optional[str] = None,
    ):
        self.config = config or GlobalConfig.read()
        current_server = self.config.get_current_server_config()
        if not current_server:
            raise ValueError("No current server config found.")
        self.api_key = api_key or current_server.api_key
        self.orign_host = current_server.server
        self.name = name
        self.namespace = namespace
        self.labels = labels
        self.model = model
        self.llms_url = f"{self.orign_host}/v1/llms"
        self.adapter = adapter

        # Parse name and namespace first
        name_parts = name.split("/")
        if len(name_parts) == 2:
            self.namespace = name_parts[0]
            self.name = name_parts[1]
        else:
            # Keep provided namespace if name doesn't contain one
            self.namespace = namespace
            self.name = name

        if not self.namespace:
            self.namespace = "-"

        logger.info(f"Using namespace: {self.namespace}, name: {self.name}")

        # Attempt to fetch the specific LLM directly
        specific_llm_url = f"{self.llms_url}/{self.namespace}/{self.name}"
        response = requests.get(
            specific_llm_url, headers={"Authorization": f"Bearer {self.api_key}"}
        )

        self.llm: Optional[V1OnlineLLM] = None
        if response.status_code == 404:
            logger.info(f"LLM {self.namespace}/{self.name} not found.")
            self.llm = None
        elif response.ok:
            self.llm = V1OnlineLLM.model_validate(response.json())
            logger.info(f"Found existing LLM {self.namespace}/{self.name}")
        else:
            # Handle other errors (e.g., 500, 403)
            response.raise_for_status()

        # If not found, create
        if not self.llm:
            request = V1OnlineLLMRequest(
                metadata=V1ResourceMetaRequest(
                    name=self.name,
                    namespace=self.namespace,
                    labels=labels,
                ),
                model=model,
                server=server,
                trainer=trainer,
                train_every=train_every,
                sample_n=sample_n,
                sample_strategy=sample_strategy,
                num_epochs=num_epochs,
            )
            logger.debug("Create LLM Request: {}", request.model_dump_json())

            create_response = requests.post(
                self.llms_url,
                json=request.model_dump(),
                headers={"Authorization": f"Bearer {self.api_key}"},
            )
            create_response.raise_for_status()

            self.llm = V1OnlineLLM.model_validate(create_response.json())
            logger.info(f"Created LLM {self.llm.metadata.name}")
        else:
            # Else, update
            logger.info(f"Found LLM {self.llm.metadata.name}, updating if necessary")
            update_request = V1UpdateOnlineLLMRequest(
                model=model,
                server=server,
                trainer=trainer,
                no_delete=no_delete,
                train_every=train_every,
                sample_n=sample_n,
                sample_strategy=sample_strategy,
                num_epochs=num_epochs,
            )
            logger.debug("Update LLM Request: {}", update_request.model_dump_json())

            patch_response = requests.patch(
                f"{self.llms_url}/{self.llm.metadata.namespace}/{self.llm.metadata.name}",
                json=update_request.model_dump(),
                headers={"Authorization": f"Bearer {self.api_key}"},
            )
            patch_response.raise_for_status()
            logger.info(f"Updated LLM {self.llm.metadata.name}")

    def generate(
        self, data: InputType | Dict[str, Any], user_key: Optional[str] = None
    ) -> OutputType:
        """
        Generate an output from the LLM.
        """
        if not self.llm or not self.llm.metadata.name:
            raise ValueError("LLM not found")

        # If the data is a ChatRequest, update its model field
        if isinstance(data, ChatRequest):
            data.model = f"{self.llm.metadata.namespace}/{self.llm.metadata.name}"

        # Handle the input data
        input_data: Dict[str, Any]
        if isinstance(data, dict):
            # Use the dictionary directly
            if "model" in data:
                data["model"] = (
                    f"{self.llm.metadata.namespace}/{self.llm.metadata.name}"
                )
            input_data = data
        elif hasattr(data, "model_dump") and callable(data.model_dump):
            # If it's a Pydantic model with model_dump method
            input_data = data.model_dump()
        else:
            # Handle unexpected input types
            raise TypeError(f"Input must be a dict or Pydantic model, got {type(data)}")

        url = f"{self.llms_url}/{self.llm.metadata.namespace}/{self.llm.metadata.name}/generate"

        logger.debug("Input data for generation: {}", input_data)

        request = V1GenerateRequest(content=input_data, user_key=user_key)

        response = requests.post(
            url,
            json=request.model_dump(),
            headers={"Authorization": f"Bearer {self.api_key}"},
        )
        response.raise_for_status()
        response_json = response.json()
        logger.debug("LLM generation response: {}", response_json)
        adapter = TypeAdapter(self.output_model)
        logger.debug("Output adapter created: {}", adapter)

        # Rebuild the adapter
        adapter.rebuild()

        # Extract the content field if it exists in the response
        if isinstance(response_json, dict) and "content" in response_json:
            content = response_json["content"]
            logger.debug("Extracted content from response: {}", content)
            val = adapter.validate_python(content)
            logger.debug("Validated output (from content): {}", val)
            return val
        else:
            val = adapter.validate_python(response_json)
            logger.debug("Validated output (from full response): {}", val)
            return val

    def train(
        self,
        wait: bool = False,
        strategy: Optional[str] = None,
        n: Optional[int] = None,
        extra_args: Optional[Dict[str, Any]] = None,
    ) -> dict:
        """
        Train the LLM.

        Args:
            wait: Whether to wait for training to complete
            strategy: Optional sample strategy
            n: Optional sample size
            extra_args: Optional additional arguments for training
        """
        if not self.llm or not self.llm.metadata.name:
            raise ValueError("LLM not found")

        url = f"{self.llms_url}/{self.llm.metadata.namespace}/{self.llm.metadata.name}/train"

        # Create request body
        request_body = {}
        if strategy is not None:
            request_body["strategy"] = strategy
        if n is not None:
            request_body["n"] = n
        if extra_args is not None:
            request_body["extra_args"] = extra_args

        response = requests.post(
            url, json=request_body, headers={"Authorization": f"Bearer {self.api_key}"}
        )
        response.raise_for_status()
        # {
        #     "success": true,
        #     "stream_id": stream_id,
        #     "message_id": message.id
        # }
        message_id = response.json()["message_id"]
        logger.info("Training started with message_id: {}", message_id)
        if wait:
            if not self.namespace:
                raise ValueError("Namespace not found")
            ref = V1ResourceReference(
                namespace=self.namespace,
                name=self.name,
                kind="Adapter",
            )
            while True:
                trainings = Training.get(
                    adapter_ref=ref,
                    labels={"message_id": message_id},
                )
                if trainings:
                    training = trainings[0]
                    if training.status == V1TrainingStatus.COMPLETED:
                        logger.info("Training completed!")
                        break
                    else:
                        logger.info(f"Training status: {training.status.value}")
                else:
                    logger.info("Waiting for training to start...")
                time.sleep(5)
        return response.json()

    def learn(
        self,
        examples: Union[Dict[str, Any], List[Dict[str, Any]], List[ExampleType]],
        train: bool = False,
    ):
        """
        Learn from a list of examples.

        Examples can be:
        - A single dictionary representing a conversation.
        - A list of dictionaries, each representing a conversation.
        - An instance of the ExampleType model.
        """

        if not self.llm or not self.llm.metadata.name:
            raise ValueError("LLM not found")

        processed_examples: List[Dict[str, Any]] = []

        if isinstance(examples, dict):
            processed_examples = [examples]
        elif isinstance(examples, list):  # type: ignore
            # Process each item in the list - could be Dict or ExampleType
            for item in examples:
                if isinstance(item, dict):
                    processed_examples.append(item)
                elif hasattr(item, "model_dump") and callable(item.model_dump):
                    processed_examples.append(item.model_dump())
                else:
                    raise TypeError(f"Unsupported type in list: {type(item).__name__}")
        elif hasattr(examples, "model_dump") and callable(examples.model_dump):
            # If it's a Pydantic model with model_dump method
            processed_examples = [examples.model_dump()]
        else:
            # Raise an error for unexpected types
            raise TypeError(f"Unsupported type for examples: {type(examples).__name__}")

        url = f"{self.llms_url}/{self.llm.metadata.namespace}/{self.llm.metadata.name}/learn"

        logger.debug("Processed example for learning: {}", processed_examples[0])
        # Now processed_examples is guaranteed to be List[Dict[str, Any]]
        request = V1ReplayBufferData(examples=processed_examples, train=train)

        response = requests.post(
            url,
            json=request.model_dump(),
            headers={"Authorization": f"Bearer {self.api_key}"},
        )

        response.raise_for_status()
        return response.json()

    @classmethod
    def load(
        cls,
        name: str,
        namespace: Optional[str] = None,
        config: Optional[GlobalConfig] = None,
    ):
        """
        Get an LLM from the remote server.
        """
        llms = cls.get(namespace=namespace, name=name, config=config)
        if not llms:
            raise ValueError("LLM not found")
        llm_v1 = llms[0]

        out = cls.__new__(cls)
        out.llm = llm_v1
        out.config = config or GlobalConfig.read()
        current_server = out.config.get_current_server_config()
        if not current_server:
            raise ValueError("No current server config found.")
        out.api_key = current_server.api_key
        out.orign_host = current_server.server
        out.llms_url = f"{out.orign_host}/v1/llms"
        out.name = name
        out.namespace = namespace
        out.model = llm_v1.model
        return out

    @classmethod
    def get(
        cls,
        name: Optional[str] = None,
        namespace: Optional[str] = None,
        config: Optional[GlobalConfig] = None,
        api_key: Optional[str] = None,
    ) -> List[V1OnlineLLM]:
        """
        Get a list of LLMs that match the optional name and/or namespace filters.
        """
        config = config or GlobalConfig.read()
        current_server = config.get_current_server_config()
        if not current_server:
            raise ValueError("No current server config found.")
        api_key = api_key or current_server.api_key
        llms_url = f"{current_server.server}/v1/llms"

        response = requests.get(
            llms_url, headers={"Authorization": f"Bearer {api_key}"}
        )
        response.raise_for_status()

        llms_response = V1OnlineLLMs.model_validate(response.json())
        filtered_llms = llms_response.llms

        if name:
            filtered_llms = [llm for llm in filtered_llms if llm.metadata.name == name]
        if namespace:
            filtered_llms = [
                llm for llm in filtered_llms if llm.metadata.namespace == namespace
            ]

        return filtered_llms

    def delete(self):
        """
        Delete the LLM.
        """
        if not self.llm or not self.llm.metadata.name:
            raise ValueError("LLM not found")

        url = f"{self.llms_url}/{self.llm.metadata.namespace}/{self.llm.metadata.name}"
        response = requests.delete(
            url, headers={"Authorization": f"Bearer {self.api_key}"}
        )
        response.raise_for_status()
        return

    def status(self) -> V1OnlineLLMStatus:
        """
        Get the status of the LLM.
        """
        if not self.llm or not self.llm.metadata.name:
            raise ValueError("LLM not found")

        llms = self.get(
            namespace=self.llm.metadata.namespace, name=self.llm.metadata.name
        )
        if not llms:
            raise ValueError("LLM not found")
        llm = llms[0]

        return llm.status

    def ref(self) -> str:
        """
        Get the resource ref for the container.
        """
        return f"{self.name}.{self.namespace}.Container"



from typing import Any, Dict, List, Optional

from nebu.containers.models import (
    V1ResourceMeta,
    V1ResourceMetaRequest,
)
from nebu.meta import V1ResourceReference
from pydantic import BaseModel, Field


class V1Trainer(BaseModel):
    processor: V1ResourceReference
    model: Optional[str] = None
    adapter: Optional[str] = None
    args: Optional[Dict[str, Any]] = None


class V1ReplayBufferRequest(BaseModel):
    metadata: V1ResourceMetaRequest
    train_every: Optional[int] = None
    sample_n: Optional[int] = Field(default=100)
    sample_strategy: Optional[str] = Field(default="Random")
    trainer: Optional[V1Trainer] = None
    num_epochs: Optional[int] = Field(default=1)


class V1UpdateReplayBufferRequest(BaseModel):
    train_every: Optional[int] = None
    sample_n: Optional[int] = None
    sample_strategy: Optional[str] = None
    trainer: Optional[V1Trainer] = None
    num_epochs: Optional[int] = None


class V1ReplayBufferStatus(BaseModel):
    num_records: Optional[int] = None
    train_idx: Optional[int] = None
    num_train_jobs: Optional[int] = None
    last_train_job: Optional[str] = None
    num_epochs: Optional[int] = None


class V1ReplayBuffer(BaseModel):
    metadata: V1ResourceMeta
    train_every: Optional[int] = None
    sample_n: Optional[int] = None
    sample_strategy: Optional[str] = None
    status: V1ReplayBufferStatus
    trainer: Optional[V1Trainer] = None
    num_epochs: Optional[int] = None


class V1ReplayBuffersResponse(BaseModel):
    buffers: List[V1ReplayBuffer]


class V1ReplayBufferData(BaseModel):
    examples: List[Dict[str, Any]]
    train: Optional[bool] = None


class V1SampleResponse(BaseModel):
    dataset_uri: Optional[str] = None
    samples: Optional[List[Dict[str, Any]]] = None


class V1SampleBufferQuery(BaseModel):
    n: int = 100
    strategy: str = "Random"
    link: bool = False


class TrainerMixin(BaseModel):
    model: str
    dataset: str
    adapter: str



import os
import time
import uuid
from dataclasses import dataclass
from typing import Any, List, Optional

from chatmux.convert import oai_to_qwen
from chatmux.openai import (
    ChatRequest,
    ChatResponse,
    CompletionChoice,
    Logprobs,
    ResponseMessage,
)
from nebu import (
    Bucket,
    ContainerConfig,
    Message,
    Processor,
    V1EnvVar,
    is_allowed,
    processor,
)
from nebu.config import GlobalConfig as NebuGlobalConfig

from orign import Adapter

setup_script = """
apt update
apt install -y git
pip install torch torchvision torchaudio --index-url https://download.pytorch.org/whl/cu126
pip uninstall -y xformers
pip3 install -U xformers --index-url https://download.pytorch.org/whl/cu126
pip install trl peft transformers bitsandbytes sentencepiece accelerate tiktoken qwen-vl-utils chatmux orign
pip install -e git+https://github.com/pbarker/unsloth-zoo.git#egg=unsloth_zoo
pip install -e git+https://github.com/pbarker/unsloth.git#egg=unsloth
"""

BASE_MODEL_ID = os.getenv("BASE_MODEL_ID", "unsloth/Qwen2.5-VL-32B-Instruct")


def init():
    import gc
    import os

    from unsloth import FastVisionModel  # type: ignore # isort: skip
    import torch  # type: ignore
    from nebu import Cache  # type: ignore

    from orign import V1Adapter

    if "state" in globals():  # <-- already loaded by an earlier worker
        print("state already loaded by an earlier worker")
        return

    gc.collect()
    torch.cuda.empty_cache()

    # os.environ.setdefault("MAX_PIXELS", "100352")

    @dataclass
    class InferenceState:
        base_model: FastVisionModel
        model_processor: Any
        base_model_id: str
        adapters: List[V1Adapter]
        cache: Cache

    print("loading model...")
    print("--- nvidia-smi before load ---")
    os.system("nvidia-smi")
    print("--- end nvidia-smi before load ---")
    time_start_load = time.time()
    base_model, model_processor = FastVisionModel.from_pretrained(
        BASE_MODEL_ID,
        load_in_4bit=False,
        # use_fast=True,
        dtype=torch.bfloat16,
        max_seq_length=32_768,
    )
    print(f"Loaded model in {time.time() - time_start_load} seconds")
    print("--- nvidia-smi after load ---")
    os.system("nvidia-smi")
    print("--- end nvidia-smi after load ---")

    global state
    state = InferenceState(
        base_model=base_model,
        model_processor=model_processor,
        base_model_id=BASE_MODEL_ID,
        adapters=[],
        cache=Cache(),
    )


def infer_qwen_vl(
    message: Message[ChatRequest],
) -> ChatResponse:
    full_time = time.time()
    from qwen_vl_utils import process_vision_info  # type: ignore
    from unsloth import FastVisionModel  # type: ignore

    global state

    print("message", message)
    training_request = message.content
    if not training_request:
        raise ValueError("No training request provided")

    # print("content", message.content)

    container_config = ContainerConfig.from_env()
    print("container_config", container_config)

    content = message.content
    if not content:
        raise ValueError("No content provided")

    load_adapter = content.model != "" and content.model != BASE_MODEL_ID
    print("load_adapter", load_adapter)

    if load_adapter:
        adapter_hot_start = time.time()

        model_parts = content.model.split("/")
        if len(model_parts) == 2:
            namespace = model_parts[0]
            name = model_parts[1]
        else:
            namespace = message.handle
            name = model_parts[0]

        print("checking for adapter", f"'{namespace}/{name}'")
        adapters = Adapter.get(namespace=namespace, name=name, api_key=message.api_key)
        if adapters:
            adapter_to_load = adapters[0]
            print("found adapter info:", adapter_to_load)

            if not is_allowed(
                adapter_to_load.metadata.owner, message.user_id, message.orgs
            ):
                raise ValueError("You are not allowed to use this adapter")

            if not adapter_to_load.base_model == BASE_MODEL_ID:
                raise ValueError(
                    "The base model of the adapter does not match the model you are trying to use"
                )

            # Check if the correct version of the adapter is already loaded
            needs_loading = True
            existing_adapter_to_unload = None
            existing_adapter_index = -1

            for idx, loaded_adapter in enumerate(state.adapters):
                print("checking against cached adapter: ", loaded_adapter)
                if (
                    loaded_adapter.metadata.name == adapter_to_load.metadata.name
                    and loaded_adapter.metadata.namespace
                    == adapter_to_load.metadata.namespace
                ):
                    if (
                        loaded_adapter.metadata.updated_at
                        == adapter_to_load.metadata.updated_at
                    ):
                        # Correct version already loaded
                        print(
                            f"Adapter version {content.model} (updated_at: {adapter_to_load.metadata.updated_at}) already loaded."
                        )
                        needs_loading = False
                        break
                    else:
                        # Different version loaded, mark for unloading
                        print(
                            f"Found different version of adapter {content.model}. Current: updated_at={loaded_adapter.metadata.updated_at}, Requested: updated_at={adapter_to_load.metadata.updated_at}"
                        )
                        existing_adapter_to_unload = loaded_adapter
                        existing_adapter_index = idx
                        break  # Stop checking, we found the one to replace

            print(
                f"Adapter hot start check took: {time.time() - adapter_hot_start} seconds"
            )

            if needs_loading:
                # Unload existing adapter if a different version was found
                if existing_adapter_to_unload:
                    try:
                        print(
                            f"Unloading existing adapter: {existing_adapter_to_unload.metadata.name} (updated_at: {existing_adapter_to_unload.metadata.updated_at})"
                        )

                        try:
                            print("peft_config: ", state.base_model.peft_config.keys())
                            # Use delete_adapter to remove it from the base model's peft config
                            state.base_model.delete_adapter(content.model)
                        except Exception as e:
                            print(
                                f"Failed to delete adapter {existing_adapter_to_unload.metadata.name}: {e}"
                            )
                        # Remove from our tracked list
                        del state.adapters[existing_adapter_index]
                        print(
                            f"Successfully unloaded adapter {existing_adapter_to_unload.metadata.name}"
                        )
                        # Optional: Clean up local files for the old adapter version?
                        # shutil.rmtree(f"./adapters/{content.model}", ignore_errors=True) # Requires import shutil
                    except Exception as e:
                        print(
                            f"Failed to unload existing adapter {existing_adapter_to_unload.metadata.name}: {e}"
                        )
                        # If unloading fails, we cannot proceed with loading the new one under the same name.
                        raise ValueError(
                            f"Failed to unload existing adapter '{existing_adapter_to_unload.metadata.name}' to load the new version."
                        ) from e

                # Download and load the new adapter
                bucket = Bucket()
                adapter_path = f"./adapters/{content.model}"  # content.model is the adapter name like 'namespace/name'
                print("copying adapter", adapter_to_load.model_uri, adapter_path)

                time_start_copy = time.time()
                bucket.copy(adapter_to_load.model_uri, adapter_path)
                print(f"Copied in {time.time() - time_start_copy} seconds")

                print("loading adapter", content.model)
                time_start_load_adapter = time.time()
                state.base_model.load_adapter(
                    adapter_path,
                    adapter_name=content.model,
                    # low_cpu_mem_usage=False, # Keep this if needed, or set based on requirements
                )
                state.adapters.append(
                    adapter_to_load
                )  # Add the newly loaded adapter info
                print(
                    f"Loaded adapter {content.model} in {time.time() - time_start_load_adapter} seconds"
                )

        else:
            raise ValueError(f"Adapter '{content.model}' not found")
        print("adapter loading/checking total time: ", time.time() - adapter_hot_start)

    # Ensure peft_config exists before trying to access keys
    loaded_adapter_names = []
    if hasattr(state.base_model, "peft_config"):
        loaded_adapter_names = list(state.base_model.peft_config.keys())
    print("loaded_adapter_names: ", loaded_adapter_names)

    if load_adapter:
        # Check if the adapter we intend to use is actually in the loaded adapters list now
        if content.model not in loaded_adapter_names:
            # This case might happen if loading failed silently, or logic error
            raise RuntimeError(
                f"Adapter {content.model} was requested but is not loaded in the model."
            )

        print("setting adapter", content.model)
        state.base_model.set_adapter(content.model)
    else:
        # Ensure no adapter is active if load_adapter is False
        print("disabling adapter")
        try:
            if hasattr(state.base_model, "disable_adapter"):
                print("Disabling any active adapter.")
                state.base_model.disable_adapter()
        except Exception as e:
            # May fail if no adapter was ever loaded or already disabled
            print(f"Failed to disable adapter (might be expected): {e}")

    print("setting model for inference")
    FastVisionModel.for_inference(state.base_model)

    content_dict = content.model_dump()
    messages_oai = content_dict["messages"]
    messages = oai_to_qwen(messages_oai)

    # Preparation for inference
    # print("preparing inputs using messages: ", messages)
    inputs_start = time.time()
    text = state.model_processor.apply_chat_template(
        messages, tokenize=False, add_generation_prompt=True
    )
    # print("text: ", text)
    # print("processing vision info: ", messages)
    image_inputs, video_inputs = process_vision_info(messages)
    inputs = state.model_processor(
        text=[text],
        images=image_inputs,
        videos=video_inputs,
        padding=True,
        return_tensors="pt",
    )
    inputs = inputs.to("cuda")
    # print("inputs", inputs)
    print(f"Inputs prepared in {time.time() - inputs_start} seconds")

    # Inference: Generation of the output
    generation_start = time.time()
    generated_ids = state.base_model.generate(
        **inputs, max_new_tokens=content.max_tokens
    )
    print(f"Generation took {time.time() - generation_start} seconds")
    generated_ids_trimmed = [
        out_ids[len(in_ids) :]
        for in_ids, out_ids in zip(inputs.input_ids, generated_ids)
    ]
    output_text = state.model_processor.batch_decode(
        generated_ids_trimmed,
        skip_special_tokens=True,
        clean_up_tokenization_spaces=False,
    )
    print("output_text", output_text)
    print(f"Generation with decoding took {time.time() - generation_start} seconds")

    # Build the Pydantic model, referencing your enumerations and classes
    response = ChatResponse(
        id=str(uuid.uuid4()),
        created=int(time.time()),
        model=content.model,
        object="chat.completion",
        choices=[
            CompletionChoice(
                index=0,
                finish_reason="stop",
                message=ResponseMessage(  # type: ignore
                    role="assistant", content=output_text[0]
                ),
                logprobs=Logprobs(content=[]),
            )
        ],
        service_tier=None,
        system_fingerprint=None,
        usage=None,
    )
    print(f"Total time: {time.time() - full_time} seconds")

    return response


def QwenVLServer(
    platform: str = "runpod",
    accelerators: List[str] = ["1:A100_SXM"],
    model: str = "unsloth/Qwen2.5-VL-32B-Instruct",
    image: str = "us-docker.pkg.dev/agentsea-dev/orign/unsloth-infer:latest",  # "pytorch/pytorch:2.6.0-cuda12.6-cudnn9-devel",
    namespace: Optional[str] = None,
    env: Optional[List[V1EnvVar]] = None,
    config: Optional[NebuGlobalConfig] = None,
    hot_reload: bool = True,
    debug: bool = False,
    min_replicas: int = 1,
    max_replicas: int = 4,
) -> Processor[ChatRequest, ChatResponse]:
    if env:
        env.append(V1EnvVar(key="BASE_MODEL_ID", value=model))
    else:
        env = [
            V1EnvVar(key="BASE_MODEL_ID", value=model),
        ]
    decorate = processor(
        image=image,
        # setup_script=setup_script,
        accelerators=accelerators,
        platform=platform,
        init_func=init,
        env=env,
        namespace=namespace,
        config=config,
        hot_reload=hot_reload,
        debug=debug,
        min_replicas=min_replicas,
        max_replicas=max_replicas,
    )
    return decorate(infer_qwen_vl)



from typing import Any, Dict, Optional

from pydantic import BaseModel, Field


class Action(BaseModel):
    """An action on a tool"""

    name: str = Field(..., description="The name of the action to be performed.")
    parameters: Dict[str, Any] = Field(
        ...,
        description="A dictionary containing parameters necessary for the action, with keys as parameter names and values as parameter details.",
    )


class ActionSelection(BaseModel):
    """An action selection from the model"""

    observation: Optional[str] = Field(
        ..., description="Observations of the current state of the environment"
    )
    reason: Optional[str] = Field(
        ...,
        description="The reason why this action was chosen, explaining the logic or rationale behind the decision.",
    )
    action: Action = Field(
        ...,
        description="The action object detailing the specific action to be taken, including its name and parameters.",
    )
    expectation: Optional[str] = Field(
        ...,
        description="The expected outcome of the action e.g. 'a login page should open'",
    )



import base64
import mimetypes
import os
import re
import secrets
import string
import tempfile
from io import BytesIO
from typing import List, Sequence

from PIL import Image

STORAGE_BUCKET_ENV = "SKILL_STORAGE_BUCKET"
STORAGE_SA_JSON_ENV = "SKILL_STORAGE_SA_JSON"


def image_to_b64(img: Image.Image, image_format: str = "PNG") -> str:
    """Converts a PIL Image to a base64-encoded string with MIME type included.

    Args:
        img (Image.Image): The PIL Image object to convert.
        image_format (str): The format to use when saving the image (e.g., 'PNG', 'JPEG').

    Returns:
        str: A base64-encoded string of the image with MIME type.
    """
    buffer = BytesIO()
    img.save(buffer, format=image_format)
    image_data = buffer.getvalue()
    buffer.close()

    mime_type = f"image/{image_format.lower()}"
    base64_encoded_data = base64.b64encode(image_data).decode("utf-8")
    return f"data:{mime_type};base64,{base64_encoded_data}"


def b64_to_image(base64_str: str) -> Image.Image:
    """Converts a base64 string to a PIL Image object.

    Args:
        base64_str (str): The base64 string, potentially with MIME type as part of a data URI.

    Returns:
        Image.Image: The converted PIL Image object.
    """
    # Strip the MIME type prefix if present
    if "," in base64_str:
        base64_str = base64_str.split(",")[1]

    image_data = base64.b64decode(base64_str)
    image = Image.open(BytesIO(image_data))
    return image


def parse_image_data(image_data_str: str):
    """Parses the image data URL to extract the MIME type and base64 data."""
    data_url_pattern = re.compile(
        r"data:(?P<mime_type>[^;]+);base64,(?P<base64_data>.+)"
    )
    match = data_url_pattern.match(image_data_str)
    if not match:
        raise ValueError(f"Invalid image data format: {image_data_str}")
    mime_type = match.group("mime_type")
    base64_data = match.group("base64_data")
    return mime_type, base64_data


def generate_random_suffix(length: int = 24) -> str:
    """Generates a random suffix for the image file name."""
    return "".join(
        secrets.choice(string.ascii_letters + string.digits) for _ in range(length)
    )


def upload_image_to_gcs(image_data: bytes, mime_type: str) -> str:
    import google.api_core.exceptions as google_exceptions
    from google.cloud import storage

    """Uploads an image to Google Cloud Storage and returns the public URL."""
    sa_json = os.getenv(STORAGE_SA_JSON_ENV)
    if not sa_json:
        try:
            storage_client = storage.Client()
        except google_exceptions.ClientError:
            raise ValueError(
                f"No GCS credentials in the environment and environment variable '{STORAGE_SA_JSON_ENV}' not set."
            )
    else:
        # Check if the service account JSON is a path or a JSON string
        if sa_json.startswith("{"):
            # Assume it's a JSON string, write to a temporary file
            with tempfile.NamedTemporaryFile(delete=False, suffix=".json") as temp_file:
                temp_file.write(sa_json.encode())
                temp_file_name = temp_file.name
        else:
            # Assume it's a path to a JSON file
            temp_file_name = sa_json

        storage_client = storage.Client.from_service_account_json(temp_file_name)

    bucket_name = os.getenv(STORAGE_BUCKET_ENV)
    if not bucket_name:
        raise ValueError(f"Environment variable {STORAGE_BUCKET_ENV} not set")

    bucket = storage_client.bucket(bucket_name)

    random_suffix = generate_random_suffix()
    extension = mimetypes.guess_extension(mime_type)
    blob_name = f"images/{random_suffix}{extension}"
    blob = bucket.blob(blob_name)

    # Create a temporary file to write the image data
    with tempfile.NamedTemporaryFile(delete=False, suffix=extension) as temp_file:
        temp_file.write(image_data)
        temp_file_name = temp_file.name

    # Upload the temporary file to Google Cloud Storage
    blob.upload_from_filename(temp_file_name)
    blob.content_type = mime_type

    # Delete the temporary file
    if sa_json and sa_json.startswith("{"):
        os.remove(temp_file_name)

    return blob.public_url


def convert_images(images: Sequence[str | Image.Image | None]) -> List[str]:
    from google.cloud import storage

    if os.getenv(STORAGE_SA_JSON_ENV):
        gcs_access = True
    else:
        try:
            _ = storage.Client()
        except Exception:
            # something went wrong initializing a storage client,
            # likely because there are no credentials in the environment
            gcs_access = False
        else:
            gcs_access = True
    new_imgs: List[str] = []
    if gcs_access:
        for img in images:
            if isinstance(img, Image.Image):
                new_imgs.append(image_to_b64(img))
            elif isinstance(img, str):
                if img.startswith("data:"):
                    # print("convert_images function, uploading img to gcs", flush=True)
                    mime_type, base64_data = parse_image_data(img)
                    image_data = base64.b64decode(base64_data)
                    public_url = upload_image_to_gcs(image_data, mime_type)
                    new_imgs.append(public_url)
                elif img.startswith("https://"):
                    new_imgs.append(img)
                else:
                    # print("convert_images function, uploading img to gcs in else", flush=True)
                    loaded_img = Image.open(img)
                    b64_img = image_to_b64(loaded_img)
                    mime_type, base64_data = parse_image_data(b64_img)
                    image_data = base64.b64decode(base64_data)
                    public_url = upload_image_to_gcs(image_data, mime_type)
                    new_imgs.append(public_url)
            else:
                raise ValueError("unnknown image type")
    else:
        for img in images:
            if img is None:  # Skip if the image is None
                continue
            # print("convert_images function, proceeding with non gcs key path", flush=True)
            if isinstance(img, Image.Image):
                new_imgs.append(image_to_b64(img))
            elif img.startswith("data:") or img.startswith("https://"):
                new_imgs.append(img)
            else:
                loaded_img = Image.open(img)
                b64_img = image_to_b64(loaded_img)
                new_imgs.append(b64_img)

    return new_imgs
